#!/usr/bin/env python
#Once upon a time...

import pdb,sys,os
import time
import re
import parabam
import pysam

import Queue as Queue2

from collections import namedtuple
from itertools import izip

from pprint import pprint as ppr


class MergePackage(parabam.core.Package):
    def __init__(self,object results,str subset_type):
        super(MergePackage,self).__init__(results)
        self.subset_type = subset_type

class Handler(parabam.core.Handler):

    def __init__(self,object parent_bam, object output_paths,object inqu,
                      object constants,object pause_qus,dict out_qu_dict):

        super(Handler,self).__init__(parent_bam,output_paths,inqu,
                                        constants,pause_qus,
                                        out_qu_dict,report=False)
        
        self._subsets = list(constants.subsets)

        #CONSTANTS
        self._CLUMP_THRESH = 1000
        self._EOF_SIGNATURE = '\x1F\x8B\x08\x04\x00\x00\x00\x00\x00\xFF'+\
                              '\x06\x00\x42\x43\x02\x00\x1B\x00\x03\x00'+\
                              '\x00\x00\x00\x00\x00\x00\x00\x00'
        self._Clump = namedtuple("Clump","sequence_ids merge_tuples")                              

        #VARIABLES

        self._out_file_objects = self.__get_out_file_objects__()
        self._subset_has_header = \
                self.__get_subset_dict__(payload=(lambda:False))
        self._merged = 0

        self._preserve_sequence = not self._constants.pair_process
        self._waiting_files = self.__get_subset_dict__(payload=dict)
        self._merge_stage = self.__get_subset_dict__()
        self._previous_merge = -1
        self._out_of_sequence_package = -2

        self._total_clumps = 0

        self._header_path = self.__get_header_sam_path__()

    def __get_header_sam_path__(self):
        header = pysam.view(*["-H",self._parent_bam.filename])
        time_of_creation = int(time.time())
        header_path = os.path.join(self._constants.temp_dir,
                                   "header_%d.bam" % (time_of_creation,))
        with open(header_path,"w") as header_sam_file:
            header_sam_file.write("".join(header))
            header_sam_file.flush()
        return header_path

    def __get_subset_dict__(self,payload=list):
        return dict( ( (subset,payload(),) for subset in self._subsets) )

    def __get_out_file_objects__(self):
        file_objects = {}
        output_paths = self._output_paths

        file_objects = {}
        for subset in self._subsets:
            output_path = output_paths[self._parent_bam.filename][subset]
            file_objects[subset] = self.__get_bam_file_obj__(output_path)

        return file_objects

    def __get_bam_file_obj__(self,output_path):
        return open(output_path,"wb")

    def __get_header_location__(self,header_path):
        header_bytes = pysam.view(*["-Hb",header_path])
        header_str = "".join(header_bytes).encode("hex")
        return (len(header_str) / 2) - len(self._EOF_SIGNATURE)

    def __periodic_action__(self,iterations):
        if self._destroy:
            try:
                pack = self._inqu.get(True,10)
                self._inqu.put(pack)
            except Queue2.Empty:
                self._finished = True

    def __new_package_action__(self,new_package,**kwargs):
        #new_package is of type parabam.merger.MergePackage
        subset = new_package.subset_type
        self.__add_to_waiting_files__(subset,new_package.results)
        self.__process_waiting_files__(subset)

    def __add_to_waiting_files__(self,subset,results):
        for merge_count,merge_path,sequence_id in results:
            merge_id = sequence_id
            if merge_id == -1:
                merge_id = self._out_of_sequence_package
                self._out_of_sequence_package += (-1)

            self._merge_stage[subset].append(merge_id)
            self._waiting_files[subset][merge_id] = (merge_count,merge_path)
        self._merge_stage[subset].sort()

    def __process_waiting_files__(self,subset,force=False):
        #TODO: Better name for valid_sequence_ids
        valid_sequence_ids = self.__get_valid_sequence_ids__(subset) 
        if len(valid_sequence_ids) > 0:

            merge_tuples = [ self._waiting_files[subset][sequence_id]\
                             for sequence_id in valid_sequence_ids ]

            clumps,clumped_sequence_ids = \
                    self.__clump_merge_files__(merge_tuples,
                                               valid_sequence_ids,
                                               force)

            if len(clumps) > 0:
                for merge_count,merge_path in clumps:
                    self.__dump_to_BAM_file__(merge_path,subset)
            
                for sequence_id in clumped_sequence_ids:
                    del self._waiting_files[subset][sequence_id]
                    self._merge_stage[subset].remove(sequence_id)

            del clumps
            del valid_sequence_ids
            del clumped_sequence_ids

    def __clump_merge_files__(self,merge_tuples,sequence_ids,force=False):
        all_above_thresh = True
        any_above_thresh = False
        total_reads = 0

        for merge_count,merge_path in merge_tuples:
            total_reads += merge_count

            if merge_count >= self._CLUMP_THRESH:
                any_above_thresh = True
            else:
                all_above_thresh = False

        if all_above_thresh:
            #no clump required
            return merge_tuples,list(sequence_ids)
        elif (total_reads < self._CLUMP_THRESH and not force) \
            or (force and total_reads == 0):
            #clump not possible
            return [],[]
        elif not any_above_thresh or force:
            #one large clump
            return [ self.__get_clump__(merge_tuples) ],list(sequence_ids)
        else:
            #several clumps
            return self.__create_clumps__(merge_tuples,list(sequence_ids))
    
    def __create_clumps__(self,merge_tuples,sequence_ids):
        clumped_sequence_ids = []
        new_merge_tuple = []

        current_clump = self.__new_clump__()
        count = 0

        for (merge_count,merge_path),sequence_id in \
                                            izip(merge_tuples,sequence_ids):
            if count > self._CLUMP_THRESH:
                new_merge_tuple.append(self.__get_clump__(\
                                            current_clump.merge_tuples))
                clumped_sequence_ids.extend(current_clump.sequence_ids)
                current_clump = self.__new_clump__()
                count = 0

            current_clump.merge_tuples.append((merge_count,merge_path))
            current_clump.sequence_ids.append(sequence_id)
            count += merge_count

        return new_merge_tuple,clumped_sequence_ids

    def __new_clump__(self):
        return self._Clump(sequence_ids=[],merge_tuples=[])

    def __get_clump__(self,merge_tuples):

        # This function may seem overly complicated but it is 
        # neccessary to handle input to pysam.cat in this way.
        #
        # The function protects pysam.cat from catting too many
        # files at once and from catting only a single file
        if len(merge_tuples) == 1:
            return merge_tuples[0]
        else:

            def get_tuple_generator(merge_tuples):
                for count,path in merge_tuples:
                    yield count,path
                return

            new_tuples = []
            total_clumped = 0
            merge_tuples_generator = get_tuple_generator(merge_tuples)

            while total_clumped < len(merge_tuples):

                clump_path = self.__get_clump_path__()
                clump_count = 0
                
                remove_paths = []
                cat_paths = []

                for count,path in merge_tuples_generator:
                    total_clumped += 1
                    remove_paths.append(path)
                    if count > 0:
                        cat_paths.append(path)
                        clump_count += count

                    if len(cat_paths) >= 250:
                        break
                

                if len(cat_paths) > 1:
                    cat_paths.insert(0,"-o%s" % (clump_path,))
                    cat_paths.insert(0,"-h%s" % (self._header_path,))
                    pysam.cat(*cat_paths)
                    new_tuples.append((clump_count,clump_path,))

                elif len(cat_paths) == 1:
                    # Rare case where a sub-clump was comprised of empty BAM 
                    # files apart from one BAM file with reads

                    new_tuples.append((clump_count,cat_paths[0]),)
                    remove_paths.remove(cat_paths[0])
                

                for path in remove_paths:
                    os.remove(path)

            return self.__get_clump__(new_tuples)

    def __get_clump_path__(self):
        clump_path = os.path.join(self._constants.temp_dir,
                              "merge_temp_%d-%d.bam" % (self._total_clumps,
                                                         time.time(),))
        self._total_clumps += 1
        return clump_path

    def __get_valid_sequence_ids__(self,subset):
        if not self._preserve_sequence:
            return self._merge_stage[subset]
        else:
            ready_to_merge = []
            prev = self._previous_merge

            self._merge_stage[subset].sort()
            for sequence_id in self._merge_stage[subset]:
                
                if sequence_id-prev > 1:
                    break
                else:
                    ready_to_merge.append(sequence_id)
                    prev = sequence_id

            if len(ready_to_merge) > 0:
                self._previous_merge = ready_to_merge[-1]

            return ready_to_merge

    def __dump_to_BAM_file__(self,merge_path,subset):
        with open(merge_path,"rb") as merge_file:
            if self._subset_has_header[subset]:
                merge_file.seek(self.__get_header_location__(merge_path))
            else:
                self._subset_has_header[subset] = True

            for binary_data in \
                self.__get_binary_from_file__(merge_file):

                self._out_file_objects[subset].write(binary_data)
                self._out_file_objects[subset].flush()
        os.remove(merge_path)

    def __get_binary_from_file__(self,open_file):
        prev_data = ""
        i = 0

        while True:
            binary_data = open_file.read(2**19)
            if not i == 0:
                combined = prev_data+binary_data
                if combined[len(combined)-28:] == self._EOF_SIGNATURE:
                    yield combined[:len(combined)-28]
                    return
                else:
                    yield prev_data

            prev_data = binary_data
            i += 1

    def __close_all_out_files__(self):
        for subset,file_obj in self._out_file_objects.items():
            file_obj.write(self._EOF_SIGNATURE)
            file_obj.flush()
            file_obj.close()
        del self._out_file_objects

    def __handler_exit__(self,**kwargs):
        for subset in self._subsets:
            self.__process_waiting_files__(subset,force=True)
        self.__close_all_out_files__()
        
#...happily ever after



import time
import sys
import os
import gc
import shutil
import gzip

import pysam
import parabam

import Queue as Queue2

from itertools import izip
from multiprocessing import Queue,Process
from abc import ABCMeta, abstractmethod

#TODO: error handle incorrect return from user rule

class Task(parabam.core.Task):
    __metaclass__=ABCMeta

    def __init__(self,parent_bam,inqu,outqu,statusqu,task_size,constants):
        super(Task, self).__init__(parent_bam=parent_bam,
                                    inqu=inqu,
                                    outqu=outqu,
                                    statusqu=statusqu,
                                    task_size=task_size,
                                    constants=constants)

        self._user_rule = constants.rule
        self._user_constants = constants.constants

    @abstractmethod
    def __handle_rule_output__(self,rule_output,read):
        pass

    def __pre_run_routine__(self,iterator,**kwargs):
        self._system = {}

    def __post_run_routine__(self,**kwargs):
        pass

    def __process_task_set__(self,iterator):
        user_rule = self._user_rule
        next_read = iterator.next 
        parent_bam = self._parent_bam
        handle_output = self.__handle_rule_output__
        user_constants = self._user_constants
        
        #StopIteration caught in parabam.core.Task.run
        for i in xrange(self._task_size):    
            read = next_read()
            rule_output = user_rule(read,user_constants,parent_bam)
            handle_output(rule_output,read)

    def __get_extension__(self,path):
        root,extension = os.path.splitext(path)
        return extension

class PairTask(Task):
    __metaclass__ = ABCMeta

    def __init__(self,
                 parent_bam,
                 inqu,
                 outqu,
                 statusqu,
                 task_size,
                 constants):
        super(PairTask, self).__init__(parent_bam=parent_bam,
                                       inqu=inqu,
                                       outqu=outqu,
                                       statusqu=statusqu,
                                       task_size=task_size,
                                       constants=constants)

        if constants.include_duplicates:
            self.__read_filter__ = self.__filter__
        else:
            self.__read_filter__ = self.__filter_duplicates__

    @abstractmethod
    def __handle_rule_output__(self,rule_output,read):
        pass

    def __pre_run_routine__(self,iterator):
        super(PairTask,self).__pre_run_routine__(iterator)
        self._loners = {}

    def __post_run_routine__(self,**kwargs):
        super(PairTask,self).__post_run_routine__()
        self.__stash_loners__(self._loners)
        del self._loners

    def __process_task_set__(self,iterator):
        
        next_read = iterator.next 
        query_loners = self.__query_loners__ 
        handle_output = self.__handle_rule_output__

        user_rule = self._user_rule
        user_constants = self._user_constants
        parent_bam = self._parent_bam

        counts = self._counts
        loners = self._loners

        cdef int size = self._task_size

        #StopIteration caught in parabam.core.Task.run
        for i in xrange(size):
            read = next_read()
            read1,read2 = query_loners(read,loners)

            if read1:
                rule_output = user_rule((read1,read2), 
                                         user_constants,
                                         parent_bam)
                handle_output(rule_output, (read1, read2,))

    def __stash_loners__(self,loners):
        loner_count = 0
        loner_path = self.__get_temp_path__("chaser")
        loner_file = pysam.AlignmentFile(loner_path,"wb",
                                         header=self._parent_bam.header)

        for qname,read in loners.items():
            loner_count += 1
            loner_file.write(read)
        self._system["chaser"] = (loner_count,loner_path,)
        loner_file.close()

    def __query_loners__(self,read,loners):
        #Pair processing only handles primary pairs
        if self.__read_filter__(read):
            return None, None
        try:
            mate = loners[read.qname]
            del loners[read.qname]
            return read,mate
        except KeyError:
            loners[read.qname] = read
            return None,None

    def __filter__(self,read):
        return read.is_secondary or (read.flag & 2048 == 2048)

    def __filter_duplicates__(self,read):
        return read.is_secondary or \
                (read.flag & 2048 == 2048) \
                    or read.is_duplicate

class ByCoordTask(Task):
    __metaclass__ = ABCMeta

    def __init__(self,
                 parent_bam,
                 inqu,
                 outqu,
                 statusqu,
                 task_size,
                 constants):
        super(ByCoordTask, self).__init__(parent_bam=parent_bam,
                                    inqu=inqu,
                                    outqu=outqu,
                                    statusqu=statusqu,
                                    task_size=task_size,
                                    constants=constants)

        #TODO: Secondary alignments???
        self.__read_filter__ = self.__filter__

    @abstractmethod
    def __handle_rule_output__(self,rule_output,read):
        pass

    def __process_task_set__(self,iterator):
        user_rule = self._rule
        next_read = iterator.next 
        parent_bam = self._parent_bam

        handle_output = self.__handle_rule_output__
        user_constants = self._user_constants

        position_max = 10000

        task_pos = None
        reads = []

        self._task_size = 0
        current_positions = 0

        #StopIteration caught in parabam.core.Task.run
        while current_positions <= position_max:
            read = next_read()
            if task_pos is None:
                task_pos = read.pos

            if task_pos == read.pos:
                reads.append(read)
                
            else:
                rule_output = user_rule(reads,user_constants,parent_bam)        
                handle_output(rule_output,reads)
                self._task_size += len(reads)
                
                del reads
                reads = [read]
                task_pos = read.pos
                current_positions += 1

    def __filter__(self,read):
        return read.is_secondary or (read.flag & 2048 == 2048)

class Handler(parabam.core.Handler):
    __metaclass__=ABCMeta

    def __init__(self, 
                object parent_bam, 
                object output_paths,
                object inqu,
                object constants,
                object pause_qus,
                dict out_qu_dict,
                object report=True):
        
        super(Handler,self).__init__(parent_bam=parent_bam,
                                     output_paths=output_paths,
                                     inqu=inqu,
                                     constants=constants,
                                     pause_qus=pause_qus,
                                     out_qu_dict=out_qu_dict)

        self._system_subsets = constants.system_subsets

        #Stage stores serve as a staging area
        #for packages to be sent on to subhandlers
        self._stage_stores = {} 
        for subset in self._system_subsets:
            self._stage_stores[subset] = []

        self._out_qu_dict = out_qu_dict
        self._stagecount = 0

    #Child classes MUST call super
    def __new_package_action__(self, new_package, **kwargs):
        results = new_package.results
        self.__auto_handle__(results,self._stats)

        if "system" in results.keys():
            self.__add_system_to_stage__(results["system"])
        
    def __add_system_to_stage__(self, system_results):
        for subset,(count,path) in system_results.items():
             self._stage_stores[subset].append((count,path))

    #Child classes MUST call super
    def __periodic_action__(self, iterations):
        self.__decide_if_finished__()
        for subset in self._system_subsets:
            if self.__test_stage_store__(subset):
                self.__add_staged_system_task__(results=self._stage_stores[subset],
                                        subset_type=subset)
                self.__update_stage_store__(subset)

    def __decide_if_finished__(self):
        if self._destroy:
            has_stage_tasks = False
            for subset,staged in self._stage_stores.items():
                if len(staged) > 0:
                    has_stage_tasks = True 
                    break
            if not has_stage_tasks:
                self._finished = True
                    
    def __update_stage_store__(self, subset):
        self._stagecount += 1
        self._stage_stores[subset] = []
        gc.collect()

    def __test_stage_store__(self, subset):
        if (not self._processing) or self._destroy:
            return len(self._stage_stores[subset]) > 0
        else:
            return len(self._stage_stores[subset]) > 10
    
    def __add_staged_system_task__(self, results, subset_type):
        if subset_type == "chaser":
            res = parabam.chaser.ChaserResults(results=results,
                                chaser_type="origin")
            self._out_qu_dict["chaser"].put(res)

    #Child classes MUST call super
    def __handler_exit__(self, **kwargs):
        if self._verbose:
            self.__standard_output__("\t- Total reads processed: %d" \
                                        % (self.__total_reads__(),))

class ByCoordFileReader(parabam.core.FileReader):

    def __init__(self,str input_path,int proc_id,object outqu,
                 int task_n,object constants,object Task,
                 object pause_qu,object inqu = None):

        super(ByCoordFileReader,self).__init__(input_path=input_path,
                                                proc_id=proc_id,
                                                outqu=outqu,
                                                task_n=task_n,
                                                constants=constants,
                                                Task=Task,
                                                pause_qu=pause_qu,
                                                inqu=inqu,)
    def __bam_generator__(self,bam_file):
        cdef int proc_id = self._proc_id
        cdef int reader_n = self._reader_n
        cdef int iterations = 0
        cdef int task_size = self._task_size

        cdef int position_max = 10000
        
        parent_iter = bam_file.fetch(until_eof = True)

        one_ahead_bam = pysam.AlignmentFile(bam_file.filename,"rb")
        one_ahead_iter = one_ahead_bam.fetch(until_eof = True)
        one_ahead_iter.next()

        while True:
            try:
                if iterations % reader_n == proc_id:
                    yield iterations
                
                observed_positions = 0
                while True:

                    current_read_pos = parent_iter.next().pos
                    try:
                        one_ahead_pos = one_ahead_iter.next().pos                        
                    except StopIteration:
                        one_ahead_pos = -1

                    if not current_read_pos == one_ahead_pos:
                        observed_positions += 1
                        if observed_positions == position_max:
                            break

                iterations += 1

            except StopIteration:
                break
        return

    def __get_parent_iter__(self,parent_bam):
        return parent_bam

class Interface(parabam.core.Interface):

    __metaclass__ = ABCMeta

    def __init__(self,pair_process = False,
                      coord_process = False,
                      include_duplicates = True,
                      debug = False,
                      ensure_unique_output=False,
                      **kwargs):

        super(Interface,self).__init__(**kwargs)

        if self.cmd_run == False:
            self.pair_process = pair_process

            self.coord_process = False
            if not self.pair_process:
                self.coord_process = coord_process

            self.include_duplicates = include_duplicates
            self.debug = debug
            self.ensure_unique_output = ensure_unique_output

    def __get_module_and_vitals__(self,code_path):
        if os.getcwd not in sys.path:
            sys.path.append(os.getcwd())

        code_path = code_path.replace(".py","")

        # TODO: I've seen an error where a bug in the imported package
        #       causes parabam to throw this message. Needs further
        #       exploration
        try:
            module = __import__(code_path, fromlist=[''])
        except ImportError: 
            sys.stderr.write("[Error] parabam can't find user specified instructions\n"\
                              "\tEnsure instruction code is in current working directory\n")
            raise SystemExit

        rule = module.rule
        constants = {}
        if hasattr(module,"set_constants"):
            module.set_constants(constants)

        return module,rule,constants

    def __cmd_args_to_class_vars__(self):

        super(Interface,self).__cmd_args_to_class_vars__()
        
        self.pair_process = self.cmd_args.pair
        self.coord_process = False

        if not self.pair_process:
            self.coord_process = self.cmd_args.coord

        self.include_duplicates = self.cmd_args.d
        self.debug = self.cmd_args.debug

    @abstractmethod
    def __get_handler_bundle__(self,**kwargs):
        pass

    @abstractmethod
    def __get_destroy_handler_order__(self):
        '''Should return a list of hanlder classes in the 
        order that they are to be destroyed by the leviathon
        all hanlder class should have a corresponding bundle entry'''
        pass 

    @abstractmethod
    def __get_output_paths__(self,input_path,final_output_paths,**kwargs):
        pass

    @abstractmethod
    def __get_task_class__(self,**kwargs):
        pass

    @abstractmethod
    def __get_queue_names__(self,**kwargs):
        '''Predefine the names of queues in this implementation
        these names are converted to queue objects by the leviathon
        use these names when reference queus in processor and handler
        bundles'''
        pass

    @abstractmethod
    def __instalise_final_output__(self,**kwargs):
        '''Allow sub-classes to instalise the dictionary
        of output paths that are returned to the caller
        of the run function.
        '''
        pass

    def __get_const_args__(self,**kwargs):
        '''Any function overwriting this one MUST make a 
            super call to this function.

            i.e args = super(<class_name>,self).__get_const_args__(**kwargs)'''

        args = {}
        
        #Dump all non function variables from this interface into constants
        for attribute in dir(self):
            is_function = hasattr(getattr(self,attribute),"__call__")
            if not is_function and not attribute.startswith("_"):
                args[attribute] = getattr(self,attribute)

        for key,val in kwargs.items():
            args[key] = val

        if self.pair_process:
            args["system_subsets"] = ["chaser"]
        else:
            args["system_subsets"] = []

        return args

    def __update_final_output_paths__(self,
                                       input_path,
                                       output_paths,
                                       final_output_paths):
        final_output_paths.update(output_paths)
    
    def __output_files_to_cwd__(self,final_output_paths):
        revised_output_paths = {}
        for input_path, analyses in final_output_paths.items():
            revised_output_paths[input_path] = {}
            for analysis_name, analysis_path in analyses.items():
                head,tail = os.path.split(analysis_path)
                new_path = os.path.abspath(os.path.join(".",tail))

                real_path = self.__move_output_file__(analysis_path,new_path)
                revised_output_paths[input_path][analysis_name] = real_path

        return revised_output_paths

    def __move_output_file__(self,current_path,new_path):
        try:
            shutil.move(current_path,new_path)
            return new_path
        except shutil.Error,e: 
            root,ext = os.path.splitext( new_path )
            alternate_path = root + "_%d" % time.time() + ext

            sys.stderr.write("WARNING: Something went wrong when copying output to working directory.\n")
            sys.stdout.write("... Trying to create output using unique filename:\n")
            sys.stdout.write("\t\t%s\n" % alternate_path)

            shutil.move(current_path,alternate_path)
            return alternate_path

    def __report_file_names__(self,final_output_paths,input_paths):
        sys.stdout.write("\t- This run will output the following files:\n")
        for master_path,analyses in final_output_paths.items():
            for name,path in analyses.items():
                root,name = os.path.split(path)
                sys.stdout.write("\t\t+ %s\n" % (name,))            

    def run(self,input_paths,**kwargs):
        self.__temp_dir_instalise__()
        self.__introduce__()

        const_args = self.__get_const_args__(**kwargs)
        constants = parabam.core.Constants(**const_args)

        task_class = self.__get_task_class__(**kwargs)
        queue_names = self.__get_queue_names__(**kwargs)

        handler_order = self.__get_destroy_handler_order__()
        handler_bundle = self.__get_handler_bundle__(**kwargs)

        if constants.pair_process:
            self.__prepare_for_pair_processing__(handler_bundle,
                                                 handler_order,
                                                 queue_names,
                                                 constants,
                                                 task_class)
        
        update_interval = self.__get_update_interval__(constants.verbose)

        filereader_class = self.__get_filereader_class__(**kwargs)

        leviathon = parabam.core.Leviathan(constants,handler_bundle,
                                           handler_order,queue_names,
                                           update_interval,task_class,
                                           filereader_class)

        final_output_paths = self.__instalise_final_output__(**kwargs)

        for input_path in input_paths:
            output_paths = self.__get_output_paths__(input_path,
                                                     final_output_paths,
                                                     **kwargs)

            self.__update_final_output_paths__(input_path,output_paths,
                                               final_output_paths)

            if constants.verbose and not self.keep_in_temp: 
                self.__report_file_names__(final_output_paths,input_path)

            leviathon.run(input_path,output_paths)

        if not self.keep_in_temp:
            final_output_paths = self.__output_files_to_cwd__(final_output_paths)
        
        self.__remove_empty_entries__(final_output_paths)

        self.__goodbye__()
        self.interface_exit()

        return final_output_paths

    def __get_update_interval__(self,verbose):
        if verbose == 1: 
            return 8000
        else:
            return 200

    def __get_filereader_class__(self,**kwargs):
        if self.coord_process:
            return ByCoordFileReader
        else:
            return parabam.core.FileReader

    def __remove_empty_entries__(self,final_output_paths):
        for master_path,child_paths in final_output_paths.items():
            if len(child_paths) == 0:
                del final_output_paths[master_path]

    def __prepare_for_pair_processing__(self,handler_bundle,handler_order,
                                            queue_names,constants,Task):        
        handler_bundle[parabam.chaser.Handler] = {"inqu":"chaser",
                                                  "constants":constants,
                                                  "out_qu_dict":["main"],
                                                  "Task":Task}

        for handler_class,handler_args in handler_bundle.items():
            if issubclass(handler_class,Handler):
                handler_bundle[handler_class]["out_qu_dict"].append("chaser")
        queue_names.append("chaser")
        handler_order.insert(0,parabam.chaser.Handler)
        constants.total_procs = constants.total_procs / 2

    def default_parser(self):
        parser = super(Interface,self).default_parser()

        parser.add_argument('--rule','-i',metavar='RULE',required=True
            ,help='The file containing the rule, written in python,\n'\
            'that we wish to apply to the input BAM.')
        parser.add_argument('--input','-b',metavar='INPUT', 
                            nargs='+',required=True
            ,help='The file(s) we wish to operate on.\n'\
            ' Multiple entries should be separated by a single space')
        parser.add_argument('--debug',action="store_true",default=False,
            help="Only the first 5million reads will be processed")
        parser.add_argument('--pair',action="store_true",default=False
            ,help="A pair processor is used instead of a conventional processor")
        parser.add_argument('--coord',action="store_true",default=False
            ,help="Engines recieve a list of reads which all map to"\
                  " the same starting position.\n This mode is not"\
                  " compatible with the `--pair` option")
        parser.add_argument('-r','--region',type=str,
                                            metavar="REGION",
                                            nargs='?',
                                            default=None
            ,help="The process will be run only on reads from\n"\
            "this region. Regions should be colon separated as\n"\
            "specified by samtools (eg \'chr1:1000,5000\')")
        parser.add_argument('-d',action="store_false",default=True,
            help="parabam will not process reads marked duplicate.")

        return parser
        
    @abstractmethod
    def get_parser(self):
        pass

#...happily ever after



#Once upon a time...
import pysam
import parabam
import shutil
import time
import sys
import os
import copy
import numpy as np

from itertools import izip
from multiprocessing import Queue
from abc import ABCMeta, abstractmethod

class StatCore(object):

    def __init__(self):
        self._counts = {}
        self._local_structures = {}
        self._system = {}

    def __pre_run_routine__(self,iterator,**kwargs):
        super(StatCore,self).__pre_run_routine__(iterator)
        for name,structure in self._constants.structures.items():
            self._local_structures[name] = structure.empty_clone()
            self._counts[name] = 0

    def __get_results__(self):
        results = {}
        results["structures"] = \
                    self.__unpack_structures__(self._local_structures)
        results["counts"] = self._counts
        results["system"] = self._system
        return results

    def __post_run_routine__(self,**kwargs):
        super(StatCore,self).__post_run_routine__()
        pass

    def __handle_rule_output__(self,rule_output,read):
        if rule_output:#Allows return False
            local_structures = self._local_structures
            for name,result in rule_output.items():
                self._counts[name] += 1
                local_structures[name].add(result)

    def __unpack_structures__(self,structures):
        unpacked = []
        for name,struc in structures.items():
            unpacked.append( (name,struc.data) )
        return unpacked

class Task(StatCore,parabam.command.Task):

    def __init__(self,parent_bam,inqu,outqu,statusqu,task_size,constants):
        
        parabam.command.Task.__init__(self,parent_bam=parent_bam,
                                            inqu=inqu,
                                            outqu=outqu,
                                            statusqu=statusqu,
                                            task_size=task_size,
                                            constants=constants)
        StatCore.__init__(self)
        
class PairTask(StatCore,parabam.command.PairTask):

    def __init__(self,parent_bam,inqu,outqu,statusqu,task_size,constants):
        
        parabam.command.PairTask.__init__(self,parent_bam=parent_bam,
                                            inqu=inqu,
                                            outqu=outqu,
                                            statusqu=statusqu,
                                            task_size=task_size,
                                            constants=constants)
        StatCore.__init__(self)


class Handler(parabam.command.Handler):

    def __init__(self,object parent_bam, object output_paths,object inqu,
                object constants,object pause_qus,dict out_qu_dict,object report=True):

        super(Handler,self).__init__(parent_bam = parent_bam,output_paths = output_paths,
                                     inqu=inqu,constants=constants,pause_qus=pause_qus,
                                     out_qu_dict=out_qu_dict)

        self._final_structures = {}
        for name,struc in self._constants.structures.items():
            self._final_structures[struc.name] = struc.empty_clone()

    def __new_package_action__(self,new_package,**kwargs):
        super(Handler,self).__new_package_action__(new_package)
        results = new_package.results
        for name,data in results["structures"]:
            final_struc = self._final_structures[name]
            final_struc.merge(data)

    def __handler_exit__(self):
        super(Handler,self).__handler_exit__()
        constants = self._constants

        #Append to global csv
        if constants.numeric_names: #Check that there are global analyses
            data_str = \
                self.__get_data_str_from_names__(constants.numeric_names,
                                                  self._final_structures)
            
            with open(self._output_paths["global"]["stats"],"a") as out_object:
                out_object.write("%s%s\n" % \
                    (self._parent_bam.filename,data_str))
        
        #Output non global data
        for name,structure in self._final_structures.items():
            if structure.struc_type == np.ndarray or \
                structure.struc_type == dict:
                structure.write_to_csv(\
                    self._output_paths[self._parent_bam.filename][name])

    def __get_data_str_from_names__(self,names,structures):
        data_str = ""
        for name in names:
            if structures[name].struc_type == np.ndarray:
                continue
            cur_data = structures[name].data
            data_str += ",%.5f" % (cur_data,)
        return data_str

class UserStructure(object):

    __metaclass__ = ABCMeta

    def __init__(self,name,struc_type,store_method,data):
        self.struc_type = struc_type
        self.store_method = store_method 
        self.data = data
        self.org_data = copy.copy(data)
        self.name = name

        if store_method == "max":
            self.add = self.add_max
            self.merge = self.merge_max
        elif store_method == "min":
            self.add = self.add_min
            self.merge = self.merge_min
        else:
            self.add = self.add_cumu
            self.merge = self.merge_cumu

    def max_decision(self,result,existing):
        return max(result,existing)

    def min_decision(self,result,exisiting):
        return min(result,exisiting)

    @abstractmethod
    def empty_clone(self):
        pass

    @abstractmethod
    def add_max(self,result):
        pass

    @abstractmethod
    def add_min(self,result):
        pass

    @abstractmethod
    def add_cumu(self,result):
        pass

    @abstractmethod
    def merge_max(self,result):
        pass

    @abstractmethod
    def merge_min(self,result):
        pass

    @abstractmethod
    def merge_cumu(self,result):
        pass

class NumericStructure(UserStructure):
    def __init__(self,name,struc_type,store_method,data,log_scaling=False):
        super(NumericStructure,self).__init__(name,struc_type,store_method,data)
        self.log_scaling = log_scaling

        if store_method == "min":
            self.data = float('inf')
            self.org_data = copy.copy(float('inf'))

    def empty_clone(self):
        return NumericStructure(self.name,
                                self.struc_type,
                                self.store_method,
                                self.org_data)

    def add_cumu(self,result):
        self.data += result

    def add_max(self,result):
        self.data = self.max_decision(result,self.data)

    def add_min(self,result):
        self.data = self.min_decision(result,self.data)

    def merge_max(self,result):
        self.data = self.max_decision(self.data,result)
        del result

    def merge_min(self,result):
        self.data = self.min_decision(self.data,result)
        del result

    def merge_cumu(self,result):
        self.data += result
        del result

#TODO: This mode doesn't work at all. Probably something to do with 
#      creating an empty clone. Counts are inflated.
class CounterStructure(UserStructure):
    def __init__(self,name,struc_type,store_method,data):
        super(CounterStructure,self).__init__(name,struc_type,store_method,data)

    def empty_clone(self):
        return CounterStructure(self.name,self.struc_type,self.store_method,self.org_data)

    def add_cumu(self,result):
        for key,value in result.items():
            try:
                self.data[key] += value
            except KeyError:
                self.data[key] = value

    def add_max(self,result):
        for key,value in result.items():
            try:
                self.data[key] = max([self.data[key],value])
            except KeyError:
                self.data[key] = value

    def add_min(self,result):
        for key,value in result.items():
            try:
                self.data[key] = min([self.data[key],value])
            except KeyError:
                self.data[key] = value

    def merge_cumu(self,result):
        self.add_cumu(result)
        del result

    def merge_max(self,result):
        self.add_max(result)
        del result

    def merge_min(self,result):
        self.add_min(result)
        del result

    def write_to_csv(self,out_path):

        with open(out_path,"w") as out_file:
            for key,value in self.data.items():
                out_str = "%s,%.5f\n" % (key,value,)
                out_file.write(out_str)

class ArrayStructure(UserStructure):
    def __init__(self,name,struc_type,store_method,data):
        super(ArrayStructure,self).__init__(name,struc_type,store_method,data)

        if self.store_method == "vstack":
            self.seen = 0
            self.add = self.add_vstack
            self.merge = self.merge_vstack

    def empty_clone(self):
        return ArrayStructure(self.name,self.struc_type,self.store_method,self.org_data)

    def add_max(self,result,coords):
        existing = self.data[coords]
        self.data[coords] = self.max_decision(result,existing)

    def add_min(self,result,coords):
        existing = self.data[coords]
        self.data[coords] = self.min_decision(result,existing)

    def add_cumu(self,result):
        self.data = np.add(self.data,result)

    def add_vstack(self,result):
        if self.seen == 0:
            self.data = result
            self.seen += 1
        else:
            self.data = np.vstack((self.data,result))

    def merge_max(self,result):
        self.data = np.maximum(self.data,result)
        del result

    def merge_min(self,result):
        self.data = np.minimum(self.data,result)
        del result

    def merge_cumu(self,result):
        self.data = np.add(self.data,result)
        del result

    def merge_vstack(self,result):
        self.add_vstack(result)
        del result

    def write_to_csv(self,out_path):

        format = []
        for x in self.data[0,:]:
            type_of_x = type(x) 
            if type_of_x == str or type_of_x == np.string_:
                format.append("%s")
            else:
                format.append("%.5f")

        np.savetxt(out_path,self.data,fmt=",".join(format),delimiter=",")

class Stat(parabam.command.Interface):

    def __init__(self,**kwargs):
        super(Stat,self).__init__(instance_name = "parabam stat", **kwargs)
    
    def __setup_other_cmd_args__(self):
        # Must implement this method to satisfy
        # inheritance
        pass

    def run_cmd(self):
        module,rule,constants = \
                self.__get_module_and_vitals__(self.cmd_args.rule)

        struc_blueprint = {}
        module.get_blueprints(struc_blueprint)

        self.run(input_paths=self.cmd_args.input,
                  constants = constants,
                  rule = rule,
                  struc_blueprint = struc_blueprint,
                  fetch_region = self.cmd_args.region,
                  specified_outpath = self.cmd_args.output)

    #TODO: The specified_output currently doesn't lead anywhere
    def run(self,input_paths,
                  constants,
                  rule,
                  struc_blueprint,
                  specified_outpath=None,
                  fetch_region=None,
                  **kwargs):
                  
        ''' Docstring! '''
        args = dict(locals())
        del args["self"]

        #Prepare state structures and insert to args
        #kwargs are later used to construct the Constant file
        #passed to all the fileprocessors and handlers
        structures = self.__create_structures__(struc_blueprint)
        numeric_names = self.__get_numeric_names__(structures)

        args["structures"] = structures
        args["numeric_names"] = numeric_names

        results = super(Stat,self).run(**args)
        return results

    def __get_destroy_handler_order__(self):
        return [Handler]

    def __get_queue_names__(self,**kwargs):
        return ["main"]

    def __get_handler_bundle__(self,**kwargs):
        handler_bundle = { Handler: {"inqu":"main","out_qu_dict":[]}}
        return handler_bundle

    def __instalise_final_output__(self,numeric_names,
                                        specified_outpath,
                                        **kwargs):
        final_output = {}

        if len(numeric_names) > 0:
            global_filename = \
                        self.__get_global_output_path__(specified_outpath)            
            self.__create_global_output_file__(global_filename,numeric_names)

            final_output["global"] = {"stats": global_filename}

        return final_output

    def __get_global_output_path__(self,specified_outpath):
        
        if specified_outpath is None:
            global_filename = \
                    os.path.join(self.temp_dir,"parabam_stat_%d_%d.csv"\
                                    % (time.time(),os.getpid()))
        else:
            global_filename = specified_outpath

        return global_filename

    def __get_output_paths__(self,input_path,
                             final_output_paths,
                             structures,
                             **kwargs):
        output_paths = {input_path:{}}

        for name,structure in structures.items():
            if not issubclass(structure.__class__,NumericStructure):
                path_id,ext = os.path.splitext(os.path.basename(input_path))
                csv_path = "%s_%s.csv" % (path_id,name,)
                output_paths[input_path][name] =\
                            os.path.join(".",self.temp_dir,csv_path)

        if "global" in final_output_paths.keys():
            output_paths["global"] = final_output_paths["global"] 
        return output_paths

    def __get_task_class__(self,**kwargs):
        if self.pair_process:
            return PairTask
        else:
            return Task

    def __get_queues__(self,object constants,**kwargs):
        queues = {"main":Queue()}
        return queues

    def __create_global_output_file__(self,output_path,numeric_names):
        header = "Sample,%s\n" % (",".join(numeric_names),)
        with open(output_path,"w") as out_obj:
            out_obj.write(header)

    def __get_numeric_names__(self,structures):
        numeric_analysis = []
        for name,structure in structures.items():
            if issubclass(structure.__class__,NumericStructure):
                numeric_analysis.append(name)
        return sorted(numeric_analysis)

    def __create_structures__(self,struc_blueprint):
        structures = {}
        class_to_type_map = {int:NumericStructure,
                             float:NumericStructure,
                             np.ndarray:ArrayStructure,
                             dict:CounterStructure}
        for name,definition in struc_blueprint.items():
            definition["name"] = name
            definition["struc_type"] = type(definition["data"])
            structures[name] = \
                    class_to_type_map[definition["struc_type"]](**definition)
        return structures

    def get_parser(self):
        #argparse imported in ./interface/parabam 
        parser = self.default_parser()

        parser.add_argument('--output','-o',
                                metavar='OUTPUT', 
                                nargs='?',
                                required=False
        ,help="Specify a name for the output CSV file. If this argument is \n"\
            "not supplied, the output will take the following form:\n"\
            "parabam_stat_[UNIX_TIME].csv")

        return parser

#...happily ever after



import pysam
import parabam
import time
import sys
import os
import gc
import shutil

from itertools import izip
from multiprocessing import Queue

class SubsetCore(object):

    def __pre_run_routine__(self,iterator,**kwargs):
        super(SubsetCore,self).__pre_run_routine__(iterator)
        self._subsets = self._constants.subsets
        self._counts = {}
        self._temp_paths = {}
        self._temp_objects = {}

        for subset in self._subsets:
            self._temp_paths[subset] = self.__get_temp_path__(subset)
            self._temp_objects[subset] = \
                    self.__get_temp_object__(self._temp_paths[subset])
            self._counts[subset] = 0

    def __get_results__(self):
        results = {}
        results["temp_paths"] = self._temp_paths
        results["counts"] = self._counts
        results["system"] = self._system

        return results

    def __write_to_subset_bam__(self,subset_type,read):
        self._counts[subset_type] += 1
        self._temp_objects[subset_type].write(read)

    def __post_run_routine__(self,**kwargs):
        super(SubsetCore,self).__post_run_routine__()
        for subset,fileobj in self._temp_objects.items():
            fileobj.close()

class Task(SubsetCore,parabam.command.Task):

    def __init__(self,parent_bam,inqu,outqu,statusqu,task_size,constants,**kwargs):
        
        parabam.command.Task.__init__(self,parent_bam=parent_bam,
                                    inqu=inqu,
                                    outqu=outqu,
                                    statusqu=statusqu,
                                    task_size=task_size,
                                    constants=constants)
        SubsetCore.__init__(self)

    def __handle_rule_output__(self,rule_output,read):             
        if type(rule_output) == bool:
            if rule_output:
                self.__write_to_subset_bam__(self._constants.subsets[0],read)          
        elif type(rule_output) == list:
            for subset,cur_read in rule_output:
                self.__write_to_subset_bam__(subset,cur_read)
        elif type(rule_output) == None:
            pass

class PairTask(SubsetCore,parabam.command.PairTask):
    def __init__(self,parent_bam,inqu,outqu,statusqu,task_size,constants):
        
        parabam.command.PairTask.__init__(self,parent_bam=parent_bam,
                                                inqu=inqu,
                                                outqu=outqu,
                                                statusqu=statusqu,
                                                task_size=task_size,
                                                constants=constants)
        SubsetCore.__init__(self)

    def __handle_rule_output__(self,rule_output,read):
        for subset,cur_read in rule_output:
            self.__write_to_subset_bam__(subset,cur_read)

class ByCoordTask(SubsetCore,parabam.command.ByCoordTask):

    def __init__(self,parent_bam,inqu,outqu,statusqu,task_size,constants,**kwargs):
        
        parabam.command.ByCoordTask.__init__(self,parent_bam=parent_bam,
                                    inqu=inqu,
                                    outqu=outqu,
                                    statusqu=statusqu,
                                    task_size=task_size,
                                    constants=constants)
        SubsetCore.__init__(self)

    def __handle_rule_output__(self,rule_output,read):     
        write_to_subset = self.__write_to_subset_bam__        
        for read in rule_output:
            write_to_subset(self._constants.subsets[0],read)

class Handler(parabam.command.Handler):

    def __init__(self,object parent_bam, object output_paths,object inqu,
                object constants,object pause_qus,dict out_qu_dict):
        
        super(Handler,self).__init__(parent_bam = parent_bam,
                                        output_paths = output_paths,
                                        inqu=inqu,
                                        constants=constants,
                                        pause_qus=pause_qus,
                                        out_qu_dict=out_qu_dict)

        self._subsets = constants.subsets
        for subset in self._subsets:
            self._stage_stores[subset] = []

    def __new_package_action__(self,new_package):
        super(Handler,self).__new_package_action__(new_package)
        results = new_package.results
        for subset in self._subsets:
            self._stage_stores[subset].append(\
                                (results["counts"][subset],
                                 results["temp_paths"][subset],
                                 new_package.sequence_id,))

    def __periodic_action__(self,iterations):
        super(Handler,self).__periodic_action__(iterations)

        for subset in self._subsets:
            if self.__test_stage_store__(subset):
                self.__add_merge_task__(results=self._stage_stores[subset],subset_type=subset)
                self.__update_stage_store__(subset)

    def __handler_exit__(self, **kwargs):
        super(Handler,self).__handler_exit__()
        #Kill the handlers handler

        for subset in self._subsets:
            #merge task 
            self.__add_merge_task__(self._stage_stores[subset],subset)
        self.__write_counts_csv__()
    
    def __add_merge_task__(self,results,subset_type):
        res = parabam.merger.MergePackage(results=results,
                                          subset_type=subset_type,)
        self._out_qu_dict["merge"].put(res)

    def __write_counts_csv__(self):
        if self._constants.output_counts:
            count_path = "./subset_counts_csv_%d.csv" % (time.time(),)
            self.__standard_output__(\
                    "- Outputting counts to following file: %s" % (count_path,))

            with open(count_path, "w") as count_file:
                key_order,header = self.__get_count_header__()
                count_file.write(header)
                for line in self.__generate_count_line__(key_order):
                    count_file.write(line)

    def __generate_count_line__(self,key_order):
        for source,stats in self._stats.items():
            line = ""
            line += "%s" % (source,)
            for key in key_order:
                line += ",%s" % (stats[key],)
            line += "\n"
            yield line

    def __get_count_header__(self):
        header = "Sample"
        keys = []
        for source, stats in self._stats.items():
            for key, val in stats.items():
                header += ","
                header += key
                keys.append(key)
            break
        header += "\n"
        return keys, header

class Subset(parabam.command.Interface):
    """The interface to parabam subset.
    Users will primarily make use of the ``run`` function."""

    def __init__(self, **kwargs):
        super(Subset,self).__init__(instance_name = "parabam subset", **kwargs)

    def run_cmd(self):
        module,rule,constants =\
                      self.__get_module_and_vitals__(self.cmd_args.rule)

        if hasattr(module,"get_subset_types"):
            subsets = module.get_subset_types()
        else:
            subsets = ["subset"]

        self.run(
            input_paths= self.cmd_args.input,
            constants = constants,
            rule = rule,
            subsets= subsets,
            fetch_region = self.cmd_args.region,
            output_counts= self.cmd_args.counts,
            ensure_unique_output = self.cmd_args.u)
    
    def run(self,input_paths,
            constants,
            rule,
            subsets,
            fetch_region=None,
            output_counts=False,
            ensure_unique_output=False,
            **kwargs):

        ''' Docstring! '''
        
        self.ensure_unique_output = False

        args = dict(locals())
        del args["self"]
        results = super(Subset,self).run(**args)
        return results

    def __get_queue_names__(self,**kwargs):
        queues = ["merge","main"]
        return queues

    def __get_destroy_handler_order__(self):
        return [Handler,parabam.merger.Handler]

    def __get_handler_bundle__(self,**kwargs):
        handler_bundle = {}

        #queues transformed by leviathon
        handler_bundle[Handler] = {"inqu":"main", 
                                   "out_qu_dict":["merge"]}

        handler_bundle[parabam.merger.Handler] = {"inqu":"merge",
                                                  "out_qu_dict":[]}
        return handler_bundle

    def __instalise_final_output__(self,**kwargs):
        return {}

    def __get_output_paths__(self,
                             input_path,
                             final_output_paths,
                             subsets,
                             **kwargs):
        
        output_paths = {input_path:{}}
        for salt,subset in enumerate(subsets):
            output_paths[input_path][subset] =\
                                     self.__get_path__(input_path,
                                                         subset,
                                                         salt)
        return output_paths

    def __get_path__(self,input_path,subset,salt):
        head,ext = os.path.splitext(input_path)
        head,tail = os.path.split(head)
        unique = ""

        if self.ensure_unique_output:
            unique = "_%d%d" % (time.time(),salt)

        return os.path.join(".",self.temp_dir, "%s_%s%s%s" \
                                    % (tail,subset,unique,ext))
            
    def __get_task_class__(self,**kwargs):
        if self.pair_process:
            return PairTask
        elif self.coord_process:
            return ByCoordTask
        else:
            return Task

    def get_parser(self):
        parser = self.default_parser()

        parser.add_argument('-u',action="store_true",default=False,
            help="The files will be appended with a unique identifier."\
                  " In the format <input_name>_<unique_id>_<subset_type>.bam")
        parser.add_argument('--counts',action="store_true",default=False,
            help="The amount of reads in each subset will be output as"\
                    " a .CSV alongside the .BAM file.")

        return parser 

#...happily ever after



#__init__.py
import core
import command
import merger
import command
import command.subset
import command.stat
import command.core
import chaser

from command.subset import Subset
from command.stat import Stat

from _version import __version__


import os
import time
import datetime
import sys
import Queue as Queue2
import gc
import shutil
import argparse
import tempfile
import signal
import traceback

import pysam

from itertools import izip
from multiprocessing import Process,Queue,freeze_support
from abc import ABCMeta, abstractmethod

class CmdLineInterface(object):
    def __init__(self,program_name = "parabam"):
        self.program_name = program_name

    def keyboard_handler(self,sig, frame):#Catch keyboard interupt and end processors
        sys.stdout.write("\r[ERROR] %s interrupted by user" \
                                         % (self.program_name))
        sys.exit(0)

    def die_gracefully(self, interface):
        if interface is not None:
            temp_dir = interface.get_temp_dir_path()
            if not temp_dir == "." or not temp_dir == "" or not temp_dir == "./": 
                interface.interface_exit()

    def handle(self,command_map,help_text):
        freeze_support()

        signal.signal(signal.SIGINT, self.keyboard_handler)     
        print_help_text = False

        if len(sys.argv) >= 2 and sys.argv[1] in command_map:
            
            command = sys.argv[1]
            #Remove command from arguments.
            sys.argv = sys.argv[1:] 
            interface = None
            try:
                #Load the command using the command line
                interface = command_map[command](cmd_run=True)
                interface.run_cmd()
                
            except SystemExit:
                print " "
                print "[Status] %s is quitting gracefully\n" \
                                                % (self.program_name,)
            except BaseException as exception:
                print " "
                print "[Error] %s stopped unexpecedtly, sorry!" \
                                                % (self.program_name,)

                traceback.print_exception(*sys.exc_info())
            finally:
                self.die_gracefully(interface)
        
        else: 
            if len(sys.argv) >= 2 and sys.argv[1] not in command_map:
                print "\nUnrecognised command: `%s`." % (sys.argv[1])
                print "Refer to manual below for a list of valid commands\n"
            print help_text

cdef class Handler:

    def __init__(self,
                  object parent_bam, 
                  object output_paths,
                  object inqu,
                  object constants,
                  object pause_qus,
                  dict out_qu_dict,
                  object report=True):

        self._parent_bam = parent_bam
        self._inqu = inqu
        self._pause_qus = pause_qus
        self._out_qu_dict = out_qu_dict

        self._report = report
        self._stats = {}

        self._constants = constants
        self._verbose = constants.verbose
        self._output_paths = output_paths

        self._periodic_interval = 10

        self._processing = True
        self._destroy = False
        self._finished = False
        self._last_update = False

        if constants.verbose == 1:
            self._verbose = True
            self._update_output = self.__level_1_output__
        elif constants.verbose == 2:
            #In the case verbose is simply "True" or "level 2"
            self._verbose = True
            self._update_output = self.__level_2_output__
        else:#catching False and -v0
            self._verbose = False
            self._update_output = self.__level_1_output__

    def __standard_output__(self,out_str):
        sys.stdout.write(out_str + "\n")
        sys.stdout.flush()

    def __level_1_output__(self,out_str):
        total_procd = self.__total_reads__()
        time = out_str.partition("Time: ")[2]
        sys.stdout.write("\t- Reads processed: %d Time: %s\n" \
                                                % (total_procd,time))
        sys.stdout.flush()

    def __level_2_output__(self,outstr):
        sys.stdout.write("\r" + outstr)
        sys.stdout.flush()

    #Must be overwritten if stats architecture is modififed
    def __total_reads__(self):
        if self._stats == {}:
            return 0
        return self._stats["Total"]

    def listen(self,update_interval):
        destroy = self._destroy

        cdef int iterations = 0
        cdef int start_time = time.time()
        cdef int dealt  = 0

        periodic_interval = self._periodic_interval #speedup alias
        finished = self.__is_finished__

        output_logic = self.__silence__
        if self._constants.verbose and self._report:
            output_logic = self.__verbose_output_logic__

        while not finished():
            iterations += 1
            #Listen for a process coming in...
            try:
                packages = []
                for i in xrange(20):
                    packages.append(self._inqu.get(False))
            except Queue2.Empty:
                #Queue empty. Continue with loop
                #time.sleep(.5)
                time.sleep(.01)

            for package in packages:
                if not package.results == {}:#If results are present...
                    self.__new_package_action__(package) #Handle the results
                    dealt += 1
                if type(package) == DestroyPackage:
                    self._destroy = True
                elif type(package) == EndProcPackage:
                    self._processing = False
            
            del packages
            if iterations % periodic_interval == 0: 
                self.__periodic_action__(iterations)

            output_logic(start_time,iterations,update_interval)

        self._inqu.close()
        self.__handler_exit__()

    def __verbose_output_logic__(self,start_time,iterations,update_interval):

        if iterations % update_interval == 0 and self._processing:
            outstr = self.__format_update__(start_time,
                                            iterations,
                                            update_interval)
            self._update_output(outstr)
        elif not self._processing and not self._last_update:
            if self._constants.verbose == 2:
                outstr = self.__format_update__(start_time,
                                                iterations,
                                                update_interval)
                self._update_output(outstr)
                sys.stdout.write("\n")
                sys.stdout.flush()

            self._last_update = True

    def __silence__(self,*args):
        pass

    def __is_finished__(self):
        if not self._destroy or not self._finished:
            return False
        return True

    def __format_update__(self,start_time,iterations,update_interval):
        first_char = self.__get_first_char__(iterations,update_interval)
        stats = "\t%s " % (first_char,)
        for name,value in self._stats.items():
            stats += "%s:%d " % (name,value)
        return "\r%s| Time: %d " %\
                (stats,self.__secs_since__(start_time),)

    def __get_first_char__(self,iterations,update_interval):
        first_chars = ["-","+","*","+"]
        return first_chars[(iterations/update_interval) % len(first_chars)]

    #This code is a little ugly. Essentially, given a results
    #dictionary, it will go through and create a sensible output
    #string.
    def __auto_handle__(self,results,stats):
        for key_one in results:
            if type(results[key_one]) is int:
                if key_one in stats:
                    stats[key_one] += results[key_one]
                else:
                    stats[key_one] = results[key_one]
            elif type(results[key_one]) is dict:
                for key_two in results[key_one]:
                    if type(results[key_one][key_two]) is int:
                        if key_two in stats:
                            stats[key_two] += results[key_one][key_two]
                        else:
                            stats[key_two] = results[key_one][key_two]

    def __secs_since__(self,since):
        return int(time.time() - since)

    def __periodic_action__(self,iterations):
        #Overwrite with something that needs
        #to be done occasionally.       
        pass

    def __new_package_action__(self,new_package,**kwargs):
        #Handle the output of a task. Input will always be of type
        #parabam.Package. New results action should always call
        #self.__auto_handle__(new_package.results)
        pass

    def __handler_exit__(self,**kwargs):
        #When the handler finishes, do this
        pass

class Task(Process):
    __metaclass__=ABCMeta

    def __init__(self, parent_bam, inqu, outqu, statusqu, task_size, constants):
        super(Task,self).__init__()
        self._parent_bam = parent_bam
        self._parent_path = self._parent_bam.filename
        self._header = self._parent_bam.header

        self._outqu = outqu
        self._inqu = inqu
        self._statusqu = statusqu

        self._task_size = task_size
        self._temp_dir = constants.temp_dir
        self._constants = constants 
        
        self._dealt = 0

    def run(self):

        bamfile = pysam.AlignmentFile(self._parent_path,"rb")
        iterator = bamfile.fetch(until_eof=True)
        generate_results = self.__generate_results__

        while True:
            try:
                package,sequence_id = self._inqu.get(False)
                self._idle_sent = False
                if type(package) == DestroyPackage:
                    bamfile.close()
                    del iterator
                    del bamfile
                    break

                seek = package                
                bamfile.seek(seek)
                time.sleep(.01)

                results = generate_results(iterator)
                
                self._dealt += 1
                time.sleep(0.005)

                self._outqu.put(Package(results=results,sequence_id=sequence_id))
                #tell filereader a batch of five jobs
                self._statusqu.put(1) 

            except Queue2.Empty:
                time.sleep(5)
            except StopIteration:
                results = self.__handle_stop_iteration__(seek,bamfile,iterator)
                self._dealt += 1 #Copy paste code here.
                time.sleep(0.005) #And here
                self.__post_run_routine__()
                self._outqu.put(Package(results=results,sequence_id=sequence_id))
        return

    def __generate_results__(self,iterator,**kwargs):
        self.__pre_run_routine__(iterator)
        self.__process_task_set__(iterator)
        results = self.__get_results__()
        results["Total"] = self._task_size
        self.__post_run_routine__()
        return results

    def __handle_stop_iteration__(self,seek,bamfile,iterator):
        results = self.__get_results__()
         
        cdef int count = 0 
        bamfile.seek(seek)
        try: #Count the amount of reads until end of file
            for read in iterator:
                count += 1
        except StopIteration:
            pass

        results["Total"] = count
        return results

    def __get_temp_object__(self,path):
        return pysam.AlignmentFile(path,"wb",header=self._parent_bam.header)

    def __get_temp_path__(self,identity):
        file_name = "%s_%d_%d_%s" %\
            (identity,self.pid,self._dealt, 
             os.path.split(self._parent_bam.filename)[1])

        return os.path.join(self._temp_dir,file_name)

    @abstractmethod
    def __pre_run_routine__(self,iterator,**kwargs):
        pass

    @abstractmethod
    def __process_task_set__(self,iterator,**kwargs):
        pass

    @abstractmethod
    def __get_results__(self):
        pass

    @abstractmethod
    def __post_run_routine__(self,**kwargs):
        pass

#The FileReader iterates over a BAM, subsets reads and
#then starts a parbam.Task process on the subsetted reads
class FileReader(Process):
    def __init__(self,str input_path,int proc_id,object outqu,
                 int task_n,object constants,object Task,
                 object pause_qu,object inqu = None):
        
        super(FileReader,self).__init__()

        self._input_path = input_path
        self._proc_id = proc_id
        
        self._outqu = outqu
        self._pause_qu = pause_qu

        self._Task = Task
        self._task_n = task_n

        self._constants = constants
        self._task_size = constants.task_size
        self._reader_n = constants.reader_n
        self._temp_dir = constants.temp_dir
        self._debug = constants.debug

        if not inqu:
            self._inqu = Queue()
        else:
            self._inqu = inqu
        self._active_jobs = 0
        self._active_jobs_thresh = ((self._task_n * 50) *\
                                        ((1000/self._task_size)+1))

    #Find data pertaining to assocd and all reads 
    #and divide pertaining to the chromosome that it is aligned to
    def run(self):

        parent_bam_mem_obj = ParentAlignmentFile(self._input_path)
        parent_bam = pysam.AlignmentFile(self._input_path,"rb")
        parent_iter = self.__get_parent_iter__(parent_bam)
        parent_generator = self.__bam_generator__(parent_iter)
        wait_for_pause = self.__wait_for_pause__
        check_inqu = self.__check_inqu__


        if self._debug:
            parent_generator = self.__debug_generator__(parent_iter)

        task_qu = Queue()

        tasks = [self._Task(parent_bam=parent_bam_mem_obj,
                      inqu=task_qu,
                      outqu=self._outqu,
                      statusqu = self._inqu,
                      task_size=self._task_size,
                      constants=self._constants) \
                        for i in xrange(self._task_n)]

        for task in tasks:
            task.start()

        for i,command in enumerate(parent_generator):
            wait_for_pause()
            task_qu.put( (parent_bam.tell(),command) )
            self._active_jobs += 1
            if i % 10 == 0:
                check_inqu()

        for n in xrange(self._task_n+1):
            task_qu.put( (DestroyPackage(),-1) )

        time.sleep(2)
        parent_bam.close()
        task_qu.close()
        return

    def __check_inqu__(self):

        while True:
            try:
                self._active_jobs -= self._inqu.get(False)
            except Queue2.Empty:
                break

        if self._active_jobs > self._active_jobs_thresh:
            while True:
                try: #Block until active jobs is less than limit
                    self._active_jobs -= self._inqu.get(False)
                    if self._active_jobs <= self._task_n*2:
                        break
                except Queue2.Empty:
                    self.__wait_for_pause__()
                    time.sleep(1)

    def __get_parent_iter__(self,parent_bam):
        if not self._constants.fetch_region:
            return parent_bam.fetch(until_eof=True)
        else:
            return parent_bam.fetch(region=self._constants.fetch_region)

    def __bam_generator__(self,parent_iter):
        cdef int proc_id = self._proc_id
        cdef int reader_n = self._reader_n
        cdef int iterations = 0
        cdef int task_size = self._task_size

        while True:
            try:
                if iterations % reader_n == proc_id:
                    yield iterations
                for x in xrange(task_size):
                    parent_iter.next()
                iterations += 1
            except StopIteration:
                break
        return

    def __debug_generator__(self,parent_iter):
        cdef int proc_id = self._proc_id
        cdef int reader_n = self._reader_n
        cdef int iterations = 0
        cdef int task_size = self._task_size

        while True:            
            try:
                if iterations % reader_n == proc_id:
                    yield iterations
                for x in xrange(task_size):
                    parent_iter.next()
                iterations += 1

                if iterations == 25:
                    break

            except StopIteration:
                break
        return

    def __send_ack__(self,qu):
        qu.put(2)
        time.sleep(1)

    def __wait_for_pause__(self):
        try: #pre clause for speed
            pause = self._pause_qu.get(False)
        except Queue2.Empty:
            return

        pause_qu = self._pause_qu
        while True:
            try:#get most recent signal
                attempt = pause_qu.get(False)
                pause = attempt
            except Queue2.Empty:
                break
        
        if pause == 1:
            self.__send_ack__(pause_qu)
            while True:
                try:
                    pause = pause_qu.get(False) 
                    if pause == 0: #Unpause signal recieved
                        self.__send_ack__(pause_qu)
                        return
                except Queue2.Empty:
                    time.sleep(.5)
 
class Leviathan(object):
    #Leviathon takes objects of file_readers and handlers and
    #chains them together.
    def __init__(self,object constants,dict handler_bundle,
                 list sequence_id,list queue_names,
                 int update,object Task,object FileReaderClass):

        self._constants = constants
        self._handler_bundle  = handler_bundle
        self.sequence_id = sequence_id
        self._update = update
        self._queue_names = queue_names
        self._Task = Task
        self._FileReaderClass = FileReaderClass
    
    def run(self,input_path,output_paths):
        parent = ParentAlignmentFile(input_path)

        default_qus = self.__create_queues__(self._queue_names) 
        pause_qus = self.__create_pause_qus__(self._constants.reader_n)

        handlers_objects,handler_inqus = \
                        self.__create_handlers__(self.sequence_id,
                                                self._handler_bundle,
                                                self._constants,
                                                default_qus,
                                                parent,
                                                output_paths,
                                                pause_qus)                        
        handlers = self.__get_handlers__(handlers_objects)

        task_n_list = self.__get_task_n__(self._constants,handlers)
        file_reader_bundles = self.__get_file_reader_bundles__(default_qus,
                                                                parent,
                                                                task_n_list,
                                                                self._constants,
                                                                pause_qus)
        file_readers = self.__get_file_readers__(file_reader_bundles)

        #Start file_readers
        for file_reader in file_readers:
            file_reader.start()

        #Start handlers:
        for handler in handlers:
            handler.start()

        #Wait for file_readers to finish
        for file_reader in file_readers:
            file_reader.join()

        #Inform handlers that processing has finished
        for handler,queue in izip(handlers,handler_inqus):
            queue.put(EndProcPackage())

        #Destory handlers
        for handler,queue in izip(handlers,handler_inqus):
            queue.put(DestroyPackage())
            handler.join()

        del default_qus
        del file_reader_bundles
        del file_readers
        del pause_qus
        del handler_inqus
        del handlers_objects
        del handlers

        gc.collect()

    def __create_pause_qus__(self,reader_n):
        return [Queue() for i in xrange(reader_n)]

    def __get_task_n__(self,constants,handlers):
        total_procs = constants.total_procs
        reader_n = constants.reader_n
        task_n_list = [total_procs / reader_n] * reader_n

        if not total_procs % reader_n == 0:
            for i,overflow in zip(xrange(len(task_n_list)),
                                  [1] * (total_procs % reader_n)):
                task_n_list[i] += overflow

        if any([ n == 0 for n in task_n_list]):
            task_n_list = [1] * reader_n

        return task_n_list 

    def __get_file_readers__(self,file_reader_bundles):
        file_readers = []
        for bundle in file_reader_bundles: 
            file_reader = self._FileReaderClass(**bundle)
            file_readers.append(file_reader)
        return file_readers

    def __proc_id_generator__(self,reader_n):
        for i in reversed(xrange(reader_n)):
            yield i

    def __get_file_reader_bundles__(self, 
                                     default_qus, 
                                     parent, 
                                     task_n_list, 
                                     constants, 
                                     pause_qus):
        bundles = []

        for proc_id,pause,task_n in izip(
                                self.__proc_id_generator__(constants.reader_n),
                                pause_qus,
                                task_n_list):

            current_bundle = {"input_path" :parent.filename,
                              "proc_id" :proc_id,
                              "task_n" :task_n,
                              "outqu" :default_qus["main"],
                              "constants" :constants,
                              "Task" :self._Task,
                              "pause_qu" :pause}

            bundles.append(current_bundle)

        return bundles
 
    def __create_queues__(self,queue_names):
        queues = {}
        for name in queue_names:
            queues[name] = Queue()
        return queues

    def __create_handlers__(self,sequence_id,handler_bundle,constants,
                            queues,parent_bam,output_paths,pause_qus):
        handlers = []
        handler_inqus = []

        for handler_class in sequence_id:
            handler_args = dict(handler_bundle[handler_class])

            handler_args["parent_bam"] = parent_bam
            handler_args["output_paths"] = output_paths
            handler_args["constants"] = constants

            #replace placeholder with queues
            handler_args["pause_qus"] = pause_qus
            handler_args["inqu"] = queues[handler_args["inqu"]]
            handler_args["out_qu_dict"] = dict(\
                    [(name,queues[name]) for name in handler_args["out_qu_dict"] ])

            handler_inqus.append(handler_args["inqu"])
            handlers.append(handler_class(**handler_args))

        return handlers,handler_inqus

    def __get_handlers__(self,handlers):
        handler_processes = []
        for handler in handlers:
            hpr = Process(target=handler.listen,args=(self._update,))
            handler_processes.append(hpr)
        return handler_processes

class Interface(object):
    '''Provides a conveinant way for providing an Interface to parabam
    programs. Includes default command_args and framework for
    command-line and programatic invocation.'''

    __metaclass__ = ABCMeta

    def __init__(self,
                  instance_name = "parabam",
                  temp_dir = None,
                  keep_in_temp = False,
                  task_size = 250000,
                  total_procs = 8,
                  reader_n = 2,
                  verbose = False,
                  announce = False,
                  cmd_run = False):

        self.instance_name = instance_name
        self.header_line = "-" * 77

        self.cmd_run = cmd_run
        if cmd_run:
            self.__cmd_args_to_class_vars__()

        else:
            self.task_size = task_size
            self.total_procs = total_procs
            self.reader_n = reader_n
            self.verbose = verbose
            self.keep_in_temp = keep_in_temp
            if not verbose:
                self.announce = False
            else:
                self.announce = announce

        self.temp_dir = temp_dir
        self.__temp_dir_instalise__()

    def get_temp_dir_path(self):
        return self.temp_dir

    def __temp_dir_instalise__(self):
        if self.temp_dir is None:
            self.temp_dir = self.__get_unique_tempdir__()
            self.control_temp_dir = True
        else:
            self.temp_dir = self.temp_dir
            self.control_temp_dir = False

    def __get_unique_tempdir__(self):
        sanitised_name = self.instance_name.replace(" ","_")
        prefix = "%s-" % (sanitised_name,)
        return tempfile.mkdtemp(prefix=prefix,dir=".")

    def __cmd_args_to_class_vars__(self):
        parser = self.get_parser()
        cmd_args = parser.parse_args()

        self.task_size = cmd_args.s
        self.total_procs = cmd_args.p
        self.verbose = cmd_args.v
        self.reader_n = cmd_args.f
    
        self.keep_in_temp = False
        self.announce = int(self.verbose) > 0

        self.cmd_args = cmd_args

    def __introduce__(self):
        if self.verbose and self.announce:
            intro =  "%s has started. Start Time: %s" \
                        % (self.instance_name,self.__get_date_time__())
            underline = ("-" * len(intro))

            sys.stdout.write(intro+"\n")
            sys.stdout.write(underline+"\n\n")
            sys.stdout.flush()
    
    def __get_date_time__(self):
        return datetime.datetime.fromtimestamp(
                            time.time()).strftime('%Y-%m-%d %H:%M:%S')

    def __goodbye__(self):
        if self.verbose and self.announce:
            outro = "%s has finished.  End Time: %s" \
                        % (self.instance_name,self.__get_date_time__())

            sys.stdout.write(("-" * len(outro)) + "\n")
            sys.stdout.write(outro + "\n")
            sys.stdout.flush()

    def interface_exit(self):
        if self.control_temp_dir and \
                not self.keep_in_temp and\
                not self.temp_dir is None:

            for path in os.listdir(self.temp_dir):
                remove_path = os.path.join(self.temp_dir,path)
                os.remove(remove_path)
            shutil.rmtree(self.temp_dir)
            self.temp_dir = None

    def default_parser(self):

        parser = ParabamParser(conflict_handler='resolve',
                    formatter_class=argparse.RawTextHelpFormatter)

        parser.add_argument('-p',type=int,nargs='?',default=4
            ,help=('The maximum amount of processes you wish\n' 
                   '%s to use. This should be less \n'
                   'than or equal to the amount of processor\n'
                   'cores in your machine [Default: 4].') % self.instance_name)
        parser.add_argument('-s',type=int,nargs='?',default=250000
            ,help="The amount of reads considered by each\n"\
                    "distributed task. [Default: 250000]")
        parser.add_argument('-f',type=int,metavar="READERS",nargs='?',default=1,
            help="The amount of open connections to the file being read.\n"\
                    "Conventional hard drives perform best with \n"\
                    "the default of 1. [Default: 1]")
        parser.add_argument('-v', choices=[0,1,2],default=0,type=int,
            help=("The amount of information output by the program:\n"
                 "\t0: No output [Default]\n"
                 "\t1: Total Reads Processed\n"
                 "\t2: Detailed output"))
        
        return parser

    @abstractmethod
    def run_cmd(self):
        #This is usualy just a function that
        #takes an argparse parser and turns 
        #passes the functions to the run function
        pass

    @abstractmethod
    def run(self):
        #The main run method. Should make a call to `interface_exit`
        pass

    @abstractmethod
    def get_parser(self):
        pass

class ParabamParser(argparse.ArgumentParser):
    def error(self, message):
        self.print_help()
        sys.stderr.write('\nerror: %s\n' % message)
        sys.exit(2)

class Constants(object):
    
    def __init__(self, **kwargs):

        for key, val in kwargs.items():
            setattr(self,key,val)

    def add(self,key,val):
        setattr(self,key,val)

class Package(object):
    def __init__(self,results,sequence_id=-1):
        self.results = results
        self.sequence_id = sequence_id

class DestroyPackage(Package):
    def __init__(self):
        super(DestroyPackage,self).__init__(results={})
        self.destroy = True

class EndProcPackage(Package):
    def __init__(self):
        super(EndProcPackage,self).__init__(results={})
        self.destroy = False

class ParentAlignmentFile(object):
    
    def __init__(self,path):
        has_index = os.path.exists(os.path.join("%s%s" % (path,".bai")))

        try:
            parent = pysam.AlignmentFile(path,"rb")
            self.filename = parent.filename
            self.references = parent.references
            self.header = parent.header
            self.lengths = parent.lengths

        except ValueError as e:
            sys.stdout.write("[Warning]: BAM header has an error:\n")
            sys.stdout.write("\t\t``  %s  ``\n" % (e,))
            sys.stdout.write("\tProcesseing will continue using the\n")
            sys.stdout.write("\tunverified header. Consider fixing the\n")
            sys.stdout.write("\tBAM header and restarting analysis\n")
            sys.stdout.flush()
            self.header = self.get_sanitised_header(parent)

        if has_index:
            self.nocoordinate = parent.nocoordinate
            self.nreferences = parent.nreferences
            self.unmapped = parent.unmapped
        else:
            self.mapped = 0
            self.nocoordinate = 0
            self.nreferences = 0
            self.unmapped = 0

        parent.close()

    def get_sanitised_header(self, parent):
        new_header = {}
        header_list = parent.text.split("\n")

        for header_entry in header_list:

            header_entry_type = header_entry[1:3]
            if header_entry_type in pysam.libcalignmentfile.VALID_HEADERS:

                entry_split = header_entry.split("\t")

                data_type = \
                 pysam.libcalignmentfile.VALID_HEADER_TYPES[header_entry_type]
                if not header_entry_type in new_header:
                    new_header[header_entry_type] = data_type()

                if data_type == list:
                    new_header[header_entry_type].append({})

                for field in entry_split[1:]:
                    self.handle_field(new_header, 
                                      header_entry_type, 
                                      field)      

        return new_header          

    def handle_field(self, new_header, header_entry_type, field):
        if ":" in field:
            known_fields = pysam.libcalignmentfile.KNOWN_HEADER_FIELDS

            field_type, colon,field_data = field.partition(":")
            is_valid_field = \
                field_type in known_fields[header_entry_type]

            if is_valid_field:
                field_data_type = known_fields[header_entry_type][field_type]
                try:
                    if type(new_header[header_entry_type]) == dict:
                        new_header[header_entry_type][field_type] = \
                            field_data_type(field_data)
                    elif type(new_header[header_entry_type]) == list:
                        new_header[header_entry_type][-1][field_type] = \
                            field_data_type(field_data)
                except ValueError:
                    pass

    def getrname(self,tid):
        return self.references[tid]

    def gettid(self,reference):
        for i,ref in enumerate(self.references):
            if reference == ref:
                return i
        return -1

#And they all lived happily ever after...
